"""
Out of the box sklearn models used to benchmark
the performance of custom PyTorch models.
"""

from sklearn import ensemble
from sklearn.metrics import mean_squared_error
from sklearn.neural_network import MLPRegressor


def fit_sklearn_mlp(X_train, y_train, X_test, y_test, hyperparams):
    """Fit an off the shelf sklearn MLP. Used for validating that custom
    pytorch model is competitive."""

    clf = MLPRegressor(
        hidden_layer_sizes=(
            hyperparams["model"]["layers"][1],
            hyperparams["model"]["layers"][2],
        ),
        activation=hyperparams["model"]["activations"][0],
        solver="adam",
        alpha=hyperparams["l2_decay"],
        batch_size=hyperparams["batch_size"],
        learning_rate_init=hyperparams["learning_rate"],
        tol=1e-2,
        max_iter=hyperparams["max_epochs"],
        shuffle=True,
        verbose=False,
        early_stopping=False,
        validation_fraction=0,
    )
    clf.fit(X_train, y_train)
    mse_train = mean_squared_error(y_train, clf.predict(X_train))
    mse_test = mean_squared_error(y_test, clf.predict(X_test))
    return {"model": clf, "mse_train": mse_train, "mse_test": mse_test}


def fit_sklearn_gbdt(X_train, y_train, X_test, y_test, hyperparams):
    """Fit an off the shelf sklearn GBDT. Used for validating that custom
    pytorch model is competitive."""

    params = {
        "n_estimators": 500,
        "max_depth": 4,
        "min_samples_split": 2,
        "learning_rate": 0.01,
        "loss": "ls",
    }

    clf = ensemble.GradientBoostingRegressor(**params)
    clf.fit(X_train, y_train)

    mse_train = mean_squared_error(y_train, clf.predict(X_train))
    mse_test = mean_squared_error(y_test, clf.predict(X_test))
    return {"model": clf, "mse_train": mse_train, "mse_test": mse_test}



import json
import time

import numpy as np
import pandas as pd
from sklearn import preprocessing
from sklearn.impute import SimpleImputer
import torch

from ..models import embed_dnn
from ..preprocessing import preprocessor


class BanditPredictor:
    """Class used to make predictions given a trained bandit model."""

    def __init__(
        self,
        experiment_specific_params,
        float_feature_order,
        id_feature_order,
        transforms,
        imputers,
        net,
        net_spec=None,
    ):
        self.experiment_specific_params = experiment_specific_params
        self.float_feature_order = float_feature_order
        self.id_feature_order = id_feature_order
        self.transforms = transforms
        self.imputers = imputers
        self.net = net
        self.net_spec = net_spec

        # the ordered decisions that we need to score over.
        self.decisions = []

    def transform_feature(self, vals, transformer=None, imputer=None):
        vals = vals.reshape(-1, 1)
        if imputer:
            vals = imputer.transform(vals)
        if transformer:
            vals = transformer.transform(vals)

        return vals

    def preprocess_input(self, input):
        # score all decisions so expand the input across decisions
        decision_meta = self.experiment_specific_params["features"]["decision"]
        if decision_meta["type"] == "C":
            decisions = decision_meta["possible_values"]
            expanded_input = [dict(input, **{"decision": d}) for d in decisions]
        elif decision_meta["type"] == "P":
            product_set_id = decision_meta["product_set_id"]
            product_set = self.experiment_specific_params["product_sets"][
                product_set_id
            ]
            decisions = product_set["ids"]
            expanded_input = [dict(input, **{"decision": d}) for d in decisions]

        self.decisions = decisions
        df = pd.DataFrame(expanded_input)
        float_feature_array = np.empty((len(df), 0))
        id_list_feature_array = np.empty((len(df), 0))

        for feature_name in self.float_feature_order:
            values = self.transform_feature(
                df[feature_name].values,
                self.transforms[feature_name],
                self.imputers[feature_name],
            )
            float_feature_array = np.append(float_feature_array, values, axis=1)

        for feature_name in self.id_feature_order:
            meta = self.experiment_specific_params["features"][feature_name]
            product_set_id = meta["product_set_id"]
            product_set_meta = self.experiment_specific_params["product_sets"][
                product_set_id
            ]
            if meta["use_dense"] is True and "dense" in product_set_meta:
                # if dense is true then convert ID's into their dense features
                dense = np.array(
                    [product_set_meta["dense"][str(i)] for i in df[feature_name].values]
                )
                for idx, feature in enumerate(product_set_meta["features"]):
                    vals = dense[:, idx]
                    values = self.transform_feature(
                        vals,
                        self.transforms[feature["name"]],
                        self.imputers[feature["name"]],
                    )
                    float_feature_array = np.append(float_feature_array, values, axis=1)
            else:
                # sparse id list features aren't preprocessed, instead they use an
                # embedding table which is built into the pytorch model
                values = self.transform_feature(df[feature_name].values)
                id_list_feature_array = np.append(id_list_feature_array, values, axis=1)

        return {
            "X_float": pd.DataFrame(float_feature_array),
            "X_id_list": pd.DataFrame(id_list_feature_array),
        }

    def preprocessed_input_to_pytorch(self, data):
        X, _ = preprocessor.data_to_pytorch(data)
        return X

    def predict(self, input):
        input = self.preprocess_input(input)
        pytorch_input = self.preprocessed_input_to_pytorch(input)
        scores = self.net.forward(**pytorch_input)
        return {"scores": scores.tolist(), "ids": self.decisions}

    def net_to_file(self, path):
        """
        Writing a Predictor object to file requires two files. This method
        writes the PyTorch net to file using PyTorch's built in serialization.
        """
        torch.save(self.net.state_dict(), path)

    def config_to_file(self, path):
        """
        Writing a Predictor object to file requires two files. This method writes
        the parameters to reconstruct the preprocessing & data imputation
        objects/logic to a JSON file.
        """
        output = {
            "net_spec": self.net_spec,
            "experiment_specific_params": self.experiment_specific_params,
            "float_feature_order": self.float_feature_order,
            "id_feature_order": self.id_feature_order,
            "transforms": {},
            "imputers": {},
        }

        # write the parameters of the feature transformers
        for feature_name, transform in self.transforms.items():
            if isinstance(transform, preprocessing.StandardScaler):
                spec = {
                    "name": "StandardScaler",
                    "mean": transform.mean_.tolist(),
                    "var": transform.var_.tolist(),
                    "scale": transform.scale_.tolist(),
                }
            elif isinstance(transform, preprocessing.OneHotEncoder):
                spec = {
                    "name": "OneHotEncoder",
                    "categories": [transform.categories_[0].tolist()],
                    "sparse": transform.sparse,
                }
            elif isinstance(transform, str):
                # id lists don't have transforms
                spec = None
            else:
                raise Exception(
                    f"Don't know how to serialize preprocessor of type {type(transform)}"
                )
            output["transforms"][feature_name] = spec

        # write the parameters of the feature imputers
        for feature_name, imputer in self.imputers.items():
            if isinstance(imputer, str):
                # id lists don't have imputers
                spec = None
            else:
                spec = {
                    "parameters": imputer.get_params(),
                    "statistics": imputer.statistics_.tolist(),
                }
            output["imputers"][feature_name] = spec

        with open(path, "w") as f:
            json.dump(output, f)

    @staticmethod
    def predictor_from_file(config_path, net_path):
        with open(config_path, "rb") as f:
            config_dict = json.load(f)

        # initialize the pytorch model and put it in `eval` mode
        net = embed_dnn.EmbedDnn(**config_dict["net_spec"])
        net.load_state_dict(torch.load(net_path))
        net.eval()

        # initialize transforms
        transforms = {}
        for feature_name, transform_spec in config_dict["transforms"].items():
            if transform_spec is None:
                # id lists don't have transforms
                transform = None
            elif transform_spec["name"] == "StandardScaler":
                transform = preprocessing.StandardScaler()
                transform.mean_ = np.array(transform_spec["mean"])
                transform.scale_ = np.array(transform_spec["scale"])
                transform.var_ = np.array(transform_spec["var"])
            elif transform_spec["name"] == "OneHotEncoder":
                transform = preprocessing.OneHotEncoder()
                transform.sparse = transform_spec["sparse"]
                transform.categories_ = np.array(transform_spec["categories"])
            else:
                raise Exception(
                    f"Don't know how to load transform_spec of type {transform_spec['name']}"
                )
            transforms[feature_name] = transform

        # initialize imputers
        imputers = {}
        for feature_name, imputer_spec in config_dict["imputers"].items():
            if imputer_spec is None:
                # id lists don't have imputers
                imputer = None
            else:
                imputer = SimpleImputer()
                imputer.set_params(**imputer_spec["parameters"])
                imputer.statistics_ = np.array(imputer_spec["statistics"])
            imputers[feature_name] = imputer

        return BanditPredictor(
            experiment_specific_params=config_dict["experiment_specific_params"],
            float_feature_order=config_dict["float_feature_order"],
            id_feature_order=config_dict["id_feature_order"],
            transforms=transforms,
            imputers=imputers,
            net=net,
            net_spec=config_dict["net_spec"],
        )



## How to release package to PyPi:
## 1) cd in package directory banditml/banditml_pkg/
## 2) bump version in setup.py
## 3) python setup.py sdist bdist_wheel
## 4) twine upload dist/* --skip-existing
## 5) See new version at https://pypi.org/project/banditml/


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="banditml",
    version="0.2.0",
    author="Edoardo Conti, Lionel Vital, Joseph Gilley",
    author_email="edoardo.conti@gmail.com",
    description="Portable Bandit ML code for training & serving consistency.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/banditml/banditml",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=["pandas==1.0.1", "torch==1.4.0", "scikit-learn==0.22.2.post1"],
)



import logging
import math
from typing import Dict, List

import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init


logger = logging.getLogger(__name__)


def gaussian_fill_w_gain(tensor, activation, dim_in, min_std=0.0) -> None:
    """ Gaussian initialization with gain."""
    gain = math.sqrt(2) if activation == "relu" else 1
    init.normal_(tensor, mean=0, std=max(gain * math.sqrt(1 / dim_in), min_std))


def build_embedding_spec(id_feature_order, feature_specs, product_sets):
    "Build the embedding spec up before creating a model."

    first_layer_dim_increase = 0
    # handle features that require embeddings
    embedding_info = {}
    for feature_name in id_feature_order:
        meta = feature_specs[feature_name]
        product_set_id = meta["product_set_id"]
        num_ids = len(product_sets[product_set_id]["ids"])
        # embedding size rule of thumb written by google:
        # https://developers.googleblog.com/2017/11/introducing-tensorflow-feature-columns.html
        embedding_dim = math.ceil(num_ids ** 0.25)
        first_layer_dim_increase += embedding_dim
        embedding_info[product_set_id] = {
            "num_ids": num_ids,
            "embedding_dim": embedding_dim,
        }
    return first_layer_dim_increase, embedding_info


class EmbedDnn(nn.Module):
    def __init__(
        self,
        layers,
        activations,
        use_batch_norm=False,
        min_std=0.0,
        dropout_ratio=0.0,
        use_layer_norm=False,
        feature_specs={},
        product_sets={},
        float_feature_order=[],
        id_feature_order=[],
        embedding_info={},
    ) -> None:
        super().__init__()
        self.layers: nn.ModuleList = nn.ModuleList()
        self.batch_norm_ops: nn.ModuleList = nn.ModuleList()
        self.activations = activations
        self.use_batch_norm = use_batch_norm
        self.dropout_layers: nn.ModuleList = nn.ModuleList()
        self.use_dropout = dropout_ratio > 0.0
        self.layer_norm_ops: nn.ModuleList = nn.ModuleList()
        self.use_layer_norm = use_layer_norm
        self.feature_specs = feature_specs
        self.float_feature_order = float_feature_order
        self.id_feature_order = id_feature_order
        self.embeddings = nn.ModuleList()
        self.embeddings_idx_map = {}
        self.net_spec = {}

        assert len(layers) >= 2, "Invalid layer schema {} for network".format(layers)

        # handle features that require embeddings
        for feature_name in id_feature_order:
            meta = feature_specs[feature_name]
            product_set_id = meta["product_set_id"]
            if product_set_id not in self.embeddings_idx_map:
                info = embedding_info[product_set_id]
                # add the +1 to num_ids as we need a 0 vector for the padding index
                self.embeddings_idx_map[product_set_id] = len(self.embeddings)
                self.embeddings.append(
                    nn.Embedding(
                        info["num_ids"] + 1, info["embedding_dim"], padding_idx=0
                    )
                )

        for i, layer in enumerate(layers[1:]):
            self.layers.append(nn.Linear(layers[i], layer))
            if self.use_batch_norm:
                self.batch_norm_ops.append(nn.BatchNorm1d(layers[i]))
            if self.use_layer_norm and i < len(layers) - 2:
                # LayerNorm is applied to the output of linear
                self.layer_norm_ops.append(nn.LayerNorm(layer))  # type: ignore
            if self.use_dropout and i < len(layers[1:]) - 1:
                # applying dropout to all layers except
                # the input and the last output layer
                self.dropout_layers.append(nn.Dropout(p=dropout_ratio))
            gaussian_fill_w_gain(
                self.layers[i].weight, self.activations[i], layers[i], min_std
            )
            init.constant_(self.layers[i].bias, 0)

    def forward(
        self,
        X_float: torch.FloatTensor,
        X_id_list: torch.LongTensor = None,
        X_id_list_idxs: torch.LongTensor = None,
    ) -> torch.FloatTensor:
        """ Forward pass for generic feed-forward DNNs. Assumes activation names
        are valid pytorch activation names.
        """

        id_feature_col_idx = 0
        # concat embedded feature to the float feature tensor
        for feature_name in self.id_feature_order:
            embedding_table_id = self.feature_specs[feature_name]["product_set_id"]
            embedding_table = self.embeddings[
                self.embeddings_idx_map[embedding_table_id]
            ]
            col_start = X_id_list_idxs[0][id_feature_col_idx]
            col_end = X_id_list_idxs[0][id_feature_col_idx + 1]
            embeddings = embedding_table(X_id_list[:, col_start:col_end])

            # https://datascience.stackexchange.com/questions/44635/does-sum-of-embeddings-make-sense
            # average the embeddings, but first drop the padded 0 embeddings
            # using a shape that maintains the tensor dimensions
            valid_row_mask = (embeddings.sum(dim=(2), keepdim=True) != 0).float()
            denom = valid_row_mask.sum(dim=(1, 2)).unsqueeze(dim=1)
            emedding_sum = embeddings.sum(dim=1)
            avg_embeddings = emedding_sum / denom
            # set Nan values to zero (case where there is an empty id list)
            avg_embeddings[avg_embeddings != avg_embeddings] = 0

            # concatenate the float tensor and embedded tensor
            X_float = torch.cat((X_float, avg_embeddings), dim=1)
            id_feature_col_idx += 2

        x = X_float
        for i, activation in enumerate(self.activations):
            if self.use_batch_norm:
                x = self.batch_norm_ops[i](x)
            x = self.layers[i](x)
            if self.use_layer_norm and i < len(self.layer_norm_ops):
                x = self.layer_norm_ops[i](x)
            if activation == "linear":
                pass
            elif activation == "tanh":
                x = torch.tanh(x)
            else:
                x = getattr(F, activation)(x)
            if self.use_dropout and i < len(self.dropout_layers):
                x = self.dropout_layers[i](x)
        return x



import json
from typing import Dict, List, Tuple

import numpy as np
import pandas as pd
from sklearn import preprocessing
from sklearn.impute import SimpleImputer
from sklearn.utils import shuffle
import torch
from torch.nn.utils.rnn import pad_sequence


def get_preprocess_feature_order(features_spec: Dict) -> List[str]:
    """Get order that features should be fed into the preprocessor. This will
    need to match bandit-app. For consistency, we will enforce the following
    ordering logic:

    Split features into float_features (i.e. N, C )& id_features (i.e. P)
        - 1st order by feature type ["N", "C"] & ["P"]
        - 2nd order by alphabetical on feature_name a->z
    """
    N, C, P = [], [], []
    for feature_name, meta in features_spec.items():
        if meta["type"] == "N":
            N.append(feature_name)
        elif meta["type"] == "C":
            C.append(feature_name)
        elif meta["type"] == "P":
            P.append(feature_name)
        else:
            raise Exception(f"Feature type {meta['type']} not supported.")

    N.sort()
    C.sort()
    P.sort()

    float_feature_order = N + C
    id_feature_order = P
    return float_feature_order, id_feature_order


def preprocess_feature(
    feature_name: str, feature_type: str, values: np.array
) -> Tuple[pd.DataFrame, int]:
    assert feature_type in ("N", "C")

    # convert to scikit learn expected format
    values = values.reshape(-1, 1)

    if feature_type == "N":
        # standard scaler seems to be right choice for numeric features
        # in regression problems
        # http://rajeshmahajan.com/standard-scaler-v-min-max-scaler-machine-learning/
        imputer = SimpleImputer(strategy="mean")
        values = imputer.fit_transform(values)
        preprocessor = preprocessing.StandardScaler()
        values = preprocessor.fit_transform(values)
        df = pd.DataFrame(values.squeeze(), columns=[feature_name])
    elif feature_type == "C":
        imputer = SimpleImputer(strategy="most_frequent")
        values = imputer.fit_transform(values)
        preprocessor = preprocessing.OneHotEncoder(sparse=False)
        values = preprocessor.fit_transform(values)
        preprocessor.col_names = [
            feature_name + "_" + str(i) for i in preprocessor.categories_[0]
        ]
        df = pd.DataFrame(values.squeeze(), columns=preprocessor.col_names)
    else:
        raise Exception(f"Feature type {feature_type} not supported.")

    return df, preprocessor, imputer


def preprocess_data(
    raw_data: pd.DataFrame, params: pd.DataFrame, shuffle_data=True
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    # start by randomizing the data upfront
    if shuffle_data:
        raw_data = shuffle(raw_data)

    # if context or decision is missing, this row is unsalvageble so drop it
    raw_data = raw_data.dropna(subset=["context", "decision"])

    # load the json string into json objects and expand into columns
    X = pd.json_normalize(raw_data["context"].apply(json.loads))
    X["decision"] = raw_data["decision"].values

    # fill in missing (NaN) rewards with empty metric maps
    raw_data["immediate_reward"] = raw_data["immediate_reward"].fillna("{}")
    immediate_y = pd.json_normalize(
        raw_data["immediate_reward"].apply(lambda x: json.loads(x))
    )
    immediate_y = immediate_y.fillna(0)

    raw_data["end_of_mdp_reward"] = raw_data["delayed_reward"].fillna("{}")
    X["end_of_mdp_reward"] = (
        raw_data["end_of_mdp_reward"].apply(lambda x: json.loads(x)).values
    )

    # first add all the end of mdp rewards if they are present. we assume end of
    # mdp rewards always have the same key in the reward function map - endOfMdpReward
    X["reward"] = X.apply(
        lambda row: row["end_of_mdp_reward"].get(str(row["decision"]), 0)
        * params["reward_function"]["endOfMdpReward"],
        axis=1,
    )

    # then add the linear combination of immediate rewards
    for metrics_name, series in immediate_y.iteritems():
        X["reward"] += series * params["reward_function"][metrics_name]

    reward_df = X["reward"]
    float_feature_df = pd.DataFrame()
    id_list_feature_df = pd.DataFrame()

    # order in which to preprocess features
    float_feature_order, id_feature_order = get_preprocess_feature_order(
        params["features"]
    )
    # final features names after preprocessing & expansion of categoricals
    final_float_feature_order, final_id_feature_order = [], []

    transforms, imputers = {}, {}
    for feature_name in float_feature_order:
        meta = params["features"][feature_name]
        df, preprocessor, imputer = preprocess_feature(
            feature_name, meta["type"], X[feature_name].values
        )
        final_float_feature_order.extend(df.columns)
        float_feature_df = pd.concat([float_feature_df, df], axis=1)
        transforms[feature_name] = preprocessor
        imputers[feature_name] = imputer

    for feature_name in id_feature_order:
        meta = params["features"][feature_name]
        products_set_id = meta["product_set_id"]
        product_set_meta = params["product_sets"][products_set_id]

        if meta["use_dense"] is True and "dense" in product_set_meta:
            # if dense is true then convert ID's into their dense features
            dense = np.array(
                [product_set_meta["dense"][str(int(i))] for i in X[feature_name].values]
            )
            for idx, feature in enumerate(product_set_meta["features"]):
                vals = dense[:, idx]
                df, preprocessor, imputer = preprocess_feature(
                    feature["name"], feature["type"], vals
                )
                final_float_feature_order.extend(df.columns)
                float_feature_df = pd.concat([float_feature_df, df], axis=1)
                transforms[feature["name"]] = preprocessor
                imputers[feature["name"]] = imputer
        else:
            # sparse id list features aren't preprocessed, instead they use an
            # embedding table which is built into the pytorch model
            final_id_feature_order.append(feature_name)
            id_list_feature_df[feature_name] = pd.Series(X[feature_name].values)
            transforms[feature_name] = params["features"][feature_name][
                "product_set_id"
            ]
            imputers[feature_name] = params["features"][feature_name]["product_set_id"]

    return {
        "y": reward_df,
        "X_float": float_feature_df,
        "X_id_list": id_list_feature_df,
        "transforms": transforms,
        "imputers": imputers,
        "float_feature_order": float_feature_order,
        "id_feature_order": id_feature_order,
        "final_float_feature_order": final_float_feature_order,
        "final_id_feature_order": final_id_feature_order,
    }


def data_to_pytorch(data: Dict):
    X = {}

    if len(data["X_float"]) > 0:
        X["X_float"] = torch.tensor(data["X_float"].values, dtype=torch.float32)

    y, X_id_list, X_id_list_idxs = None, None, None

    # hack needed due to skorch not handling objects in .fit() besides
    # a dict of lists or tensors (dict of dicts not supported.)
    id_list_pad_idxs = []
    pad_idx = 0

    for _, series in data["X_id_list"].iteritems():
        # wrap decisions in lists
        if series.dtype in (int, float):
            series = series.apply(lambda x: [x])

        pre_pad = [torch.tensor(i, dtype=torch.long) for i in series]
        post_pad = pad_sequence(pre_pad, batch_first=True, padding_value=0)
        idx_offset = post_pad.shape[1]
        id_list_pad_idxs.extend([pad_idx, pad_idx + idx_offset])
        pad_idx += idx_offset

        if X_id_list is not None:
            X_id_list = torch.cat((X_id_list, post_pad), dim=1)
        else:
            X_id_list = post_pad

    # skorch requires all inputs to .fit() to have the same length so we
    # need to repeat our id_list_pad_idxs list.
    if X_id_list is not None:
        X["X_id_list"] = X_id_list
        X["X_id_list_idxs"] = torch.tensor([id_list_pad_idxs for i in X_id_list])

    # this function is used to make predictions as well, where there is no y
    if "y" in data:
        y = torch.tensor(data["y"].values, dtype=torch.float32).unsqueeze(dim=1)

    return X, y
